#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 Inria

from typing import Optional, Set, Tuple

import numpy as np
from gymnasium import spaces

from upkie.utils.clamp import clamp_and_warn
from upkie.utils.robot_state import RobotState

from .upkie_base_env import UpkieBaseEnv


class UpkieServos(UpkieBaseEnv):
    r"""!
    Upkie with actions and observations corresponding to the moteus servo API.

    \anchor upkie_servos_description

    ### Action space

    The action space is a dictionary with one key for each servo:

    - `left_hip`: left hip joint (qdd100)
    - `left_hip`: left knee joint (qdd100)
    - `left_hip`: left wheel joint (mj5208)
    - `right_hip`: right hip joint (qdd100)
    - `right_hip`: right knee joint (qdd100)
    - `right_hip`: right wheel joint (mj5208)

    The value for each servo dictionary is itself a dictionary with the
    following keys:

    - `position`: commanded joint angle \f$\theta^*\f$ in [rad] (NaN to
       disable) (required).
    - `velocity`: commanded joint velocity \f$\dot{\theta}^*\f$ in [rad] /
       [s] (required).
    - `feedforward_torque`: feedforward joint torque \f$\tau_{\mathit{ff}}\f$
       in [N m].
    - `kp_scale`: scaling factor \f$k_{p}^{\mathit{scale}}\f$ applied to the
       position feedback gain, between zero and one.
    - `kd_scale`: scaling factor \f$k_{d}^{\mathit{scale}}\f$ applied to the
       velocity feedback gain, between zero and one.
    - `maximum_torque`: maximum joint torque \f$\tau_{\mathit{max}}\f$
       (feedforward + feedback) enforced during the whole actuation step, in
       [N m].

    The resulting torque applied by the servo is then:

    \f[
    \begin{align*}
    \tau & = \underset{
            [-\tau_{\mathit{max}}, +\tau_{\mathit{max}}]}{
            \mathrm{clamp}
        }
        \left(
            \tau_{\mathit{ff}} +
            k_{p} k_{p}^{\mathit{scale}} (\theta^* - \theta) +
            k_{d} k_{d}^{\mathit{scale}} (\dot{\theta}^* - \dot{\theta}))
        \right)
    \end{align*}
    \f]

    Position and velocity gains \f$k_{p}\f$ and \f$k_{d}\f$ are configured in
    each moteus controller directly and don't change during execution. We can
    rather modulate the overall feedback gains via the normalized parameters
    \f$k_{p}^{\mathit{scale}} \in [0, 1]\f$ and \f$k_{d}^{\mathit{scale}} \in
    [0, 1]\f$. Note that the servo regulates the torque above at its own
    frequency, which is higher (typically 40 kHz) than the agent and the spine
    frequencies. See the [moteus
    reference](https://github.com/mjbots/moteus/blob/13c171c697ce6f60a73c9385e6fe951957313d1d/docs/reference.md#theory-of-operation)
    for more details.

    ### Observation space

    The observation space is a dictionary with one key for each servo. The
    value for each key is a dictionary with keys:

    - `position`: Joint angle in [rad].
    - `velocity`: Joint velocity in [rad] / [s].
    - `torque`: Joint torque in [N m].
    - `temperature`: Servo temperature in degree Celsius.
    - `voltage`: Power bus voltage of the servo, in [V].

    As with all Upkie environments, full observations from the spine (detailed
    in \ref observations) are also available in the `info` dictionary
    returned by the reset and step functions.
    """

    ACTION_KEYS: Tuple[str, str, str, str, str, str] = (
        "position",
        "velocity",
        "feedforward_torque",
        "kp_scale",
        "kd_scale",
        "maximum_torque",
    )

    ACTION_MASK: Set[str] = set()

    ## \var action_space
    ## Action space.
    action_space: spaces.dict.Dict

    ## \var observation_space
    ## Observation space.
    observation_space: spaces.dict.Dict

    ## \var version
    ## Environment version number.
    version = 4

    def __init__(
        self,
        fall_pitch: float = 1.0,
        frequency: float = 200.0,
        frequency_checks: bool = True,
        init_state: Optional[RobotState] = None,
        regulate_frequency: bool = True,
        shm_name: str = "/upkie",
        spine_config: Optional[dict] = None,
    ):
        r"""!
        Initialize environment.

        \param fall_pitch Fall pitch angle, in radians.
        \param frequency Regulated frequency of the control loop, in Hz.
        \param frequency_checks If `regulate_frequency` is set and this
            parameter is true (default), a warning is issued every time the
            control loop runs slower than the desired `frequency`. Set this
            parameter to false to disable these warnings.
        \param init_state Initial state of the robot, only used in simulation.
        \param regulate_frequency Enables loop frequency regulation.
        \param shm_name Name of shared-memory file.
        \param spine_config Additional spine configuration overriding the
            default `upkie.config.SPINE_CONFIG`. The combined configuration
            dictionary is sent to the spine at every `reset`.
        """
        super().__init__(
            fall_pitch=fall_pitch,
            frequency=frequency,
            frequency_checks=frequency_checks,
            init_state=init_state,
            regulate_frequency=regulate_frequency,
            shm_name=shm_name,
            spine_config=spine_config,
        )

        action_space = {}
        neutral_action = {}
        max_action = {}
        min_action = {}
        servo_space = {}

        for joint in self.model.joints:
            action_space[joint.name] = spaces.Dict(
                {
                    "position": spaces.Box(
                        low=joint.limit.lower,
                        high=joint.limit.upper,
                        shape=(1,),
                        dtype=float,
                    ),
                    "velocity": spaces.Box(
                        low=-joint.limit.velocity,
                        high=+joint.limit.velocity,
                        shape=(1,),
                        dtype=float,
                    ),
                    "feedforward_torque": spaces.Box(
                        low=-joint.limit.effort,
                        high=+joint.limit.effort,
                        shape=(1,),
                        dtype=float,
                    ),
                    "kp_scale": spaces.Box(
                        low=0.0,
                        high=1.0,
                        shape=(1,),
                        dtype=float,
                    ),
                    "kd_scale": spaces.Box(
                        low=0.0,
                        high=1.0,
                        shape=(1,),
                        dtype=float,
                    ),
                    "maximum_torque": spaces.Box(
                        low=0.0,
                        high=joint.limit.effort,
                        shape=(1,),
                        dtype=float,
                    ),
                }
            )
            servo_space[joint.name] = spaces.Dict(
                {
                    "position": spaces.Box(
                        low=joint.limit.lower,
                        high=joint.limit.upper,
                        shape=(1,),
                        dtype=float,
                    ),
                    "velocity": spaces.Box(
                        low=-joint.limit.velocity,
                        high=+joint.limit.velocity,
                        shape=(1,),
                        dtype=float,
                    ),
                    "torque": spaces.Box(
                        low=-joint.limit.effort,
                        high=+joint.limit.effort,
                        shape=(1,),
                        dtype=float,
                    ),
                    "temperature": spaces.Box(
                        low=0.0,
                        high=100.0,
                        shape=(1,),
                        dtype=float,
                    ),
                    "voltage": spaces.Box(
                        low=10.0,  # moteus min 10 V
                        high=44.0,  # moteus max 44 V
                        shape=(1,),
                        dtype=float,
                    ),
                }
            )
            neutral_action[joint.name] = {
                "position": np.nan,
                "velocity": 0.0,
                "feedforward_torque": 0.0,
                "kp_scale": 1.0,
                "kd_scale": 1.0,
                "maximum_torque": joint.limit.effort,
            }
            max_action[joint.name] = {
                "position": joint.limit.upper,
                "velocity": joint.limit.velocity,
                "feedforward_torque": joint.limit.effort,
                "kp_scale": 1.0,
                "kd_scale": 1.0,
                "maximum_torque": joint.limit.effort,
            }
            min_action[joint.name] = {
                "position": joint.limit.lower,
                "velocity": -joint.limit.velocity,
                "feedforward_torque": -joint.limit.effort,
                "kp_scale": 0.0,
                "kd_scale": 0.0,
                "maximum_torque": 0.0,
            }

        # gymnasium.Env: action_space
        self.action_space = spaces.Dict(action_space)

        # gymnasium.Env: observation_space
        self.observation_space = spaces.Dict(servo_space)

        # Class attributes
        self.__neutral_action = neutral_action
        self.__max_action = max_action
        self.__min_action = min_action

    def get_neutral_action(self) -> dict:
        r"""!
        Get the neutral action where servos don't move.

        \return Neutral action where servos don't move.
        """
        return self.__neutral_action.copy()

    def get_env_observation(self, spine_observation: dict):
        r"""!
        Extract environment observation from spine observation dictionary.

        \param spine_observation Full observation dictionary from the spine.
        \return Environment observation.
        """
        # If creating a new object turns out to be too slow we can switch to
        # updating in-place.
        return {
            joint.name: {
                key: np.array(
                    [spine_observation["servo"][joint.name][key]],
                    dtype=float,
                )
                for key in self.observation_space[joint.name]
            }
            for joint in self.model.joints
        }

    def get_spine_action(self, env_action: dict) -> dict:
        r"""!
        Convert environment action to a spine action dictionary.

        \param env_action Environment action.
        \return Spine action dictionary.
        """
        spine_action = {"servo": {}}
        for joint in self.model.joints:
            servo_action = {}
            for key in self.ACTION_KEYS:
                action = (
                    env_action[joint.name][key]
                    if key in env_action[joint.name]
                    and (not self.ACTION_MASK or key in self.ACTION_MASK)
                    else self.__neutral_action[joint.name][key]
                )
                action_value = (
                    action.item()
                    if isinstance(action, np.ndarray)
                    else float(action)
                )
                servo_action[key] = clamp_and_warn(
                    action_value,
                    self.__min_action[joint.name][key],
                    self.__max_action[joint.name][key],
                    label=f"{joint.name}: {key}",
                )
            spine_action["servo"][joint.name] = servo_action
        return spine_action

    def get_reward(self, observation: dict, action: dict) -> float:
        r"""!
        Get reward from observation and action.

        \param observation Environment observation.
        \param action Environment action.
        \return Reward.
        """
        return 1.0



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2024 Inria
# SPDX-License-Identifier: Apache-2.0

"""Wheeled inverted pendulum."""

from typing import Any, Optional, Tuple, Union

import gymnasium
import numpy as np
import upkie_description
from gymnasium import spaces
from loop_rate_limiters import RateLimiter
from numpy import cos, sin

from upkie.exceptions import MissingOptionalDependency, UpkieRuntimeError
from upkie.model import Model
from upkie.utils.clamp import clamp_and_warn
from upkie.utils.spdlog import logging

from .rewards import WheeledInvertedPendulumReward

GRAVITY: float = 9.81  # [m] / [s]²


class WheeledInvertedPendulum(gymnasium.Env):
    r"""!
    Wheeled inverted pendulum model.

    This environment helps test and debug behaviors in a perfect or
    noise-controlled setting. It has the same observation and action spaces as
    \ref upkie.envs.upkie_ground_velocity.UpkieGroundVelocity. Model
    assumptions and discretization are summed up in [this
    note](https://scaron.info/robotics/wheeled-inverted-pendulum-model.html).

    ### Action space

    The action corresponds to the ground velocity resulting from wheel
    velocities. The action vector is simply:

    <table>
        <tr>
            <td><strong>Index</strong></td>
            <td><strong>Description</strong></td>
            </tr>
        <tr>
            <td>0</td>
            <td>Ground velocity in [m] / [s].</td>
        </tr>
    </table>

    ### Observation space

    Vectorized observations have the following structure:

    <table>
        <tr>
            <td><strong>Index</strong></td>
            <td><strong>Description</strong></td>
        </tr>
        <tr>
            <td>0</td>
            <td>Pitch angle of the base with respect to the world vertical, in
            radians. This angle is positive when the robot leans forward.</td>
        </tr>
        <tr>
            <td>1</td>
            <td>Position of the average wheel contact point, in meters.</td>
        </tr>
        <tr>
            <td>2</td>
            <td>Body angular velocity of the base frame along its lateral axis,
            in radians per seconds.</td>
        </tr>
        <tr>
            <td>3</td>
            <td>Velocity of the average wheel contact point, in meters per
            seconds.</td>
        </tr>
    </table>
    """

    ## \var action_space
    ## Action space.
    action_space: spaces.box.Box

    ## \var dt
    ## Period of the control loop in seconds.
    dt: float

    ## \var fall_pitch
    ## Fall pitch angle, in radians.
    fall_pitch: float

    ## \var length
    ## Length of the inverted pendulum.
    length: float

    ## \var metadata
    ## Metadata of the environment containing rendering modes.
    metadata = {"render_modes": ["plot"]}

    ## \var model
    ## Robot model read from its URDF description.
    model: Model

    ## \var observation_space
    ## Observation space.
    observation_space: spaces.box.Box

    ## \var plot
    ## Optional plot used for rendering.
    plot: Optional[Any]

    ## \var render_mode
    ## Inherited from gymnasium.Env.
    render_mode: Optional[str]

    ## \var reward
    ## Reward function of the environment.
    reward: WheeledInvertedPendulumReward

    ## \var uncertainty
    ## Biases and noise levels on modeled system uncertainties.
    uncertainty: "WheeledInvertedPendulum.Uncertainty"

    ## \var version
    ## Environment version number.
    version = 1

    class Uncertainty:
        r"""!
        Biases and noise levels on modeled system uncertainties.
        """

        ## \var accelerometer_bias
        ## Bias vector added to accelerometer measurements in the base frame.
        accelerometer_bias: Union[np.ndarray, float]

        ## \var accelerometer_noise
        ## Vector of standard deviations for white noise added to accelerometer
        ## measurements in the base frame.
        accelerometer_noise: Union[np.ndarray, float]

        ## \var observation_bias
        ## Bias vector added to state observations.
        observation_bias: Union[np.ndarray, float]

        ## \var observation_noise
        ## Vector of standard deviations for white noise added to state
        ## observations.
        observation_noise: Union[np.ndarray, float]

        def __init__(
            self,
            accelerometer_bias: Union[np.ndarray, float] = 0.0,
            accelerometer_noise: Union[np.ndarray, float] = 0.0,
            observation_bias: Union[np.ndarray, float] = 0.0,
            observation_noise: Union[np.ndarray, float] = 0.0,
        ):
            r"""!
            Initialize uncertainties.
            """
            self.accelerometer_bias = accelerometer_bias
            self.accelerometer_noise = accelerometer_noise
            self.observation_bias = observation_bias
            self.observation_noise = observation_noise

        def accelerometer(self) -> np.ndarray:
            r"""!
            Get accelerometer uncertainty vector.

            \return Acceleration uncertainty vector in the base frame.
            """
            return np.random.normal(
                loc=self.accelerometer_bias,
                scale=self.accelerometer_noise,
                size=(2,),
            )

        def observation(self) -> np.ndarray:
            r"""!
            Get observation uncertainty vector.
            """
            return np.random.normal(
                loc=self.observation_bias,
                scale=self.observation_noise,
                size=(4,),
            )

    def __init__(
        self,
        fall_pitch: float = 1.0,
        frequency: float = 200.0,
        frequency_checks: bool = True,
        length: float = 0.6,
        max_ground_accel: float = 10.0,
        max_ground_velocity: float = 1.0,
        regulate_frequency: bool = True,
        render_mode: Optional[str] = None,
        reward: Optional[WheeledInvertedPendulumReward] = None,
        uncertainty: Optional["WheeledInvertedPendulum.Uncertainty"] = None,
    ):
        r"""!
        Initialize a new environment.

        \param fall_pitch Fall detection pitch angle, in [rad].
        \param frequency Regulated frequency of the control loop, in Hz.
        \param frequency_checks If `regulate_frequency` is set and this
            parameter is true (default), a warning is issued every time the
            control loop runs slower than the desired `frequency`. Set this
            parameter to false to disable these warnings.
        \param length Length of the pole.
        \param max_ground_velocity Maximum commanded ground velocity in [m] /
            [s].
        \param max_ground_accel  Maximum acceleration of the ground point,
            in [m] / [s]².
        \param regulate_frequency Enables loop frequency regulation.
        \param render_mode Rendering mode, set to "plot" for live plotting.
        \param reward Reward function of the environment.
        \param uncertainty Uncertainty biases and noise magnitudes.
        """
        if (
            render_mode is not None
            and render_mode not in self.metadata["render_modes"]
        ):
            raise UpkieRuntimeError(
                f"Render mode '{render_mode}' is not in "
                f"{self.metadata['render_modes']}"
            )

        # gymnasium.Env: observation_space
        MAX_BASE_PITCH: float = np.pi
        MAX_GROUND_POSITION: float = float("inf")
        MAX_BASE_ANGULAR_VELOCITY: float = 1000.0  # rad/s
        observation_limit = np.array(
            [
                MAX_BASE_PITCH,
                MAX_GROUND_POSITION,
                MAX_BASE_ANGULAR_VELOCITY,
                max_ground_velocity,
            ],
            dtype=float,
        )
        self.observation_space = spaces.Box(
            -observation_limit,
            +observation_limit,
            shape=observation_limit.shape,
            dtype=observation_limit.dtype,
        )

        # gymnasium.Env: action_space
        action_limit = np.array([max_ground_velocity], dtype=float)
        self.action_space = spaces.Box(
            -action_limit,
            +action_limit,
            shape=action_limit.shape,
            dtype=action_limit.dtype,
        )

        reward: WheeledInvertedPendulumReward = (
            reward if reward is not None else WheeledInvertedPendulumReward()
        )

        rate = None
        dt = 1.0 / frequency
        if regulate_frequency:
            rate = RateLimiter(
                frequency,
                name=f"{self.__class__.__name__} rate limiter",
                warn=frequency_checks,
            )

        model = Model(upkie_description.URDF_PATH)
        spine_observation = {
            "base_orientation": {
                "pitch": 0.0,
                "angular_velocity": np.zeros(3),
            },
            "imu": {
                "raw_angular_velocity": np.zeros(3),
                "raw_linear_acceleration": np.zeros(3),
            },
            "servo": {
                joint.name: {
                    "position": 0.0,
                    "velocity": 0.0,
                }
                for joint in model.upper_leg_joints
            },
            "sim": {
                "base": {
                    "position": np.array([0.0, 0.0, length]),
                    "orientation": np.array([1.0, 0.0, 0.0, 0.0]),
                    "linear_velocity": np.zeros(3),
                    "angular_velocity": np.zeros(3),
                }
            },
            "wheel_odometry": {
                "position": 0.0,
                "velocity": 0.0,
            },
        }

        if uncertainty is None:
            uncertainty = WheeledInvertedPendulum.Uncertainty()

        self.__accel = np.zeros(2)
        self.__max_ground_accel = max_ground_accel
        self.__max_ground_velocity = max_ground_velocity
        self.__noise = np.zeros(4)
        self.__rate = rate
        self.__regulate_frequency = regulate_frequency
        self.__spine_observation = spine_observation
        self.__state = np.zeros(4)
        self.dt = dt
        self.fall_pitch = fall_pitch
        self.length = length
        self.model = model
        self.plot = None
        self.render_mode = render_mode
        self.reward = reward
        self.uncertainty = uncertainty

    def reset(
        self,
        *,
        seed: Optional[int] = None,
        options: Optional[dict] = None,
    ) -> Tuple[np.ndarray, dict]:
        r"""!
        Resets the spine and get an initial observation.

        \param seed Number used to initialize the environment’s internal random
            number generator.
        \param options Currently unused.
        \return
            - `observation`: Initial vectorized observation, i.e. an element
              of the environment's `observation_space`.
            - `info`: Dictionary with auxiliary diagnostic information. For
              Upkie this is the full observation dictionary sent by the spine.
        """
        super().reset(seed=seed)
        self.__state = np.zeros(4)
        observation = self.__state
        info = {"spine_observation": self._get_spine_observation()}
        if self.render_mode == "plot":
            self._reset_plot()
        return observation, info

    def _reset_plot(self):
        r"""!
        Reset plot, called when render mode is set to "plot".
        """
        if self.plot is None:
            try:
                from matplotlive import LivePlot

                self.plot = LivePlot(
                    timestep=self.dt,
                    duration=1.0,
                    ylim=(-0.01, 0.01),
                    ylim_right=(-0.5, 0.5),
                )
            except ImportError as exn:
                raise MissingOptionalDependency(
                    "matplotlive not found, run `pip install matplotlive`"
                ) from exn
        self.plot.reset()
        self.plot.add_left("pitch", "b-")
        self.plot.left_axis.set_ylabel(r"Pitch angle (rad)", color="b")
        self.plot.left_axis.tick_params(axis="y", labelcolor="b")
        self.plot.add_right("ground_position", "g-")
        self.plot.right_axis.set_ylabel(r"Position (m)", color="g")
        self.plot.right_axis.tick_params(axis="y", labelcolor="g")
        self.plot.redraw()

    @staticmethod
    def _integrate(x, v, a, dt):
        """Explicit second-order Euler integration in a vector space."""
        x2 = x + dt * (v + dt * (a / 2))
        v2 = v + dt * a
        return x2, v2

    def step(
        self,
        action: np.ndarray,
    ) -> Tuple[np.ndarray, float, bool, bool, dict]:
        r"""!
        Run one timestep of the environment's dynamics. When the end of the
        episode is reached, you are responsible for calling `reset()` to reset
        the environment's state.

        \param action Action from the agent.
        \return
            - `observation`: Observation of the environment, i.e. an element
              of its `observation_space`.
            - `reward`: Reward returned after taking the action.
            - `terminated`: Whether the agent reached a terminal state,
              which can be a good or a bad thing. When true, the user needs to
              call `reset()`.
            - `truncated`: Whether the episode is reaching max number of
              steps. This boolean can signal a premature end of the episode,
              i.e. before a terminal state is reached. When true, the user
              needs to call `reset()`.
            - `info`: Dictionary with auxiliary diagnostic information. For
              us this is the full observation dictionary coming from the spine.
        """
        if self.__rate is not None:
            self.__rate.sleep()  # wait until clock tick to send the action

        theta_0, r_0, thetad_0, rd_0 = self.__state
        rd_next = clamp_and_warn(
            action[0],
            -self.__max_ground_velocity,
            self.__max_ground_velocity,
            "ground_velocity",
        )
        rdd = clamp_and_warn(
            (rd_next - rd_0) / self.dt,
            -self.__max_ground_accel,
            self.__max_ground_accel,
            "ground_acceleration",
        )
        thetadd = (GRAVITY * sin(theta_0) - rdd * cos(theta_0)) / self.length
        r, rd = self._integrate(r_0, rd_0, rdd, self.dt)
        theta, thetad = self._integrate(theta_0, thetad_0, thetadd, self.dt)

        self.__state[0] = theta
        self.__state[1] = r
        self.__state[2] = thetad
        self.__state[3] = rd
        self.__accel[0] = thetadd
        self.__accel[1] = rdd

        if self.render_mode == "plot":
            self._render_plot()

        observation = self.__state + self.uncertainty.observation()
        reward = self.reward(
            pitch=theta,
            ground_position=r,
            angular_velocity=thetad,
            ground_velocity=rd,
        )
        terminated = self.detect_fall(theta)
        truncated = False
        info = {"spine_observation": self._get_spine_observation()}
        return observation, reward, terminated, truncated, info

    def detect_fall(self, pitch: float) -> bool:
        r"""!
        Detect a fall based on the base-to-world pitch angle.

        \param pitch Pitch angle of the rotation from base to world frame.
        \return True if and only if a fall is detected.
        """
        if abs(pitch) > self.fall_pitch:
            logging.warning(
                "Fall detected (pitch=%.2f rad, fall_pitch=%.2f rad)",
                abs(pitch),
                self.fall_pitch,
            )
            return True
        return False

    def render(self):
        r"""!
        Render state to a live plot.
        """
        if self.render_mode == "plot":
            return self._render_plot()

    def _render_plot(self):
        r"""!
        Render plot.
        """
        self.plot.send("pitch", self.__state[0])
        self.plot.send("ground_position", self.__state[1])
        self.plot.update()

    def _get_imu_acceleration_in_base(
        self,
        state: Optional[np.ndarray] = None,
        accel: Optional[np.ndarray] = None,
    ) -> np.ndarray:
        r"""!
        Get IMU acceleration in the base frame.

        \param state State vector.
        \param accel Acceleration vector.
        \return IMU acceleration in the base frame.
        """
        theta, r, thetad, rd = state if state is not None else self.__state
        thetadd, rdd = accel if accel is not None else self.__accel
        rotation_base_to_world = np.array(
            [[np.cos(theta), np.sin(theta)], [-np.sin(theta), np.cos(theta)]]
        )
        u = np.array([rdd, +GRAVITY])
        v = self.length * np.array([thetadd, -(thetad**2)])
        proper_accel_in_world = u + rotation_base_to_world @ v
        rotation_world_to_base = rotation_base_to_world.T
        proper_accel_in_base = rotation_world_to_base @ proper_accel_in_world
        uncertainty_in_base = self.uncertainty.accelerometer()
        return proper_accel_in_base + uncertainty_in_base

    def _get_spine_observation(self):
        r"""!
        Get spine observation.
        """
        linear_to_3d = np.array([[1.0, 0.0], [0.0, 0.0], [0.0, 1.0]])
        theta, r, thetad, rd = self.__state
        accel_2d = self._get_imu_acceleration_in_base()
        accel_in_base = linear_to_3d @ accel_2d
        omega_in_base = np.array([0.0, thetad, 0.0])
        accel_in_imu = self.model.rotation_base_to_imu @ accel_in_base
        omega_in_imu = self.model.rotation_base_to_imu @ omega_in_base

        e = np.array([np.sin(theta), np.cos(theta)])
        e_orth = np.array([np.cos(theta), -np.sin(theta)])
        p = np.array([r, 0.0]) + self.length * e
        v = np.array([rd, 0.0]) + self.length * thetad * e_orth

        obs = self.__spine_observation  # reference, not a copy
        obs["sim"]["base"]["position"] = linear_to_3d @ p
        obs["sim"]["base"]["linear_velocity"] = linear_to_3d @ v
        obs["base_orientation"]["angular_velocity"][1] = thetad
        obs["base_orientation"]["pitch"] = theta
        obs["imu"]["raw_angular_velocity"] = omega_in_imu
        obs["imu"]["raw_linear_acceleration"] = accel_in_imu
        obs["wheel_odometry"]["position"] = r
        obs["wheel_odometry"]["velocity"] = rd
        return obs

    def _get_state(self) -> np.ndarray:
        r"""!
        Get a copy of the current internal state.
        """
        return self.__state.copy()

    def log(self, name: str, entry: Any):
        r"""!
        Log a new entry to the "log" key of the action dictionary.

        \param name Name of the entry.
        \param entry Dictionary to log along with the actual action.
        """
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron

## \namespace upkie.envs
## \brief Gymnasium environments.
##
## See the [list of environments](\ref environments) for more details.

import gymnasium as gym

from .upkie_base_env import UpkieBaseEnv
from .upkie_ground_velocity import UpkieGroundVelocity
from .upkie_servo_positions import UpkieServoPositions
from .upkie_servo_torques import UpkieServoTorques
from .upkie_servos import UpkieServos
from .wheeled_inverted_pendulum import WheeledInvertedPendulum


def register() -> None:
    """!
    Register Upkie environments with Gymnasium.
    """
    envs = (
        ("UpkieGroundVelocity", UpkieGroundVelocity),
        ("UpkieServoPositions", UpkieServoPositions),
        ("UpkieServoTorques", UpkieServoTorques),
        ("UpkieServos", UpkieServos),
        ("WheeledInvertedPendulum", WheeledInvertedPendulum),
    )
    for env_name, env_class in envs:
        gym.envs.registration.register(
            id=f"{env_name}-v{env_class.version}",
            entry_point=f"upkie.envs:{env_name}",
        )


__all__ = [
    "UpkieBaseEnv",
    "UpkieGroundVelocity",
    "UpkieServoPositions",
    "UpkieServoTorques",
    "UpkieServos",
    "WheeledInvertedPendulum",
    "register",
]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron
# Copyright 2023 Inria

import math
from typing import Dict, Optional, Tuple

import numpy as np
from gymnasium import spaces

from upkie.exceptions import UpkieException
from upkie.utils.clamp import clamp_and_warn
from upkie.utils.filters import low_pass_filter
from upkie.utils.robot_state import RobotState

from .rewards import WheeledInvertedPendulumReward
from .upkie_base_env import UpkieBaseEnv


class UpkieGroundVelocity(UpkieBaseEnv):
    r"""!
    Environment where Upkie is used as a wheeled inverted pendulum.

    \anchor upkie_ground_velocity_description

    With this environment, Upkie keeps its legs straight and actions only
    affect wheel velocities. This way, it behaves like a <a
    href="https://scaron.info/robotics/wheeled-inverted-pendulum-model.html">wheeled
    inverted pendulum</a>. This ground-velocity environment is used for
    instance by the [MPC balancer](https://github.com/upkie/mpc_balancer/) and
    [PPO balancer](https://github.com/upkie/ppo_balancer) agents.

    \note For reinforcement learning with neural networks: the observation
    space and action space are not normalized.

    ### Action space

    The action corresponds to the ground velocity resulting from wheel
    velocities. The action vector is simply:

    \f[
    a =\begin{bmatrix} \dot{p}^* \end{bmatrix}
    \f]

    where we denote by \f$\dot{p}^*\f$ the commanded ground velocity in [m] /
    [s], which is internally converted into wheel velocity commands. Note that,
    while this action is not normalized, [-1, 1] m/s is a reasonable range for
    ground velocities.

    ### Observation space

    Vectorized observations have the following structure:

    \f[
    \begin{align*}
    o &= \begin{bmatrix} \theta \\ p \\ \dot{\theta} \\ \dot{p} \end{bmatrix}
    \end{align*}
    \f]

    where we denote by:

    - \f$\theta\f$ the pitch angle of the base with respect to the world
      vertical, in radians. This angle is positive when the robot leans
      forward.
    - \f$p\f$ the position of the average wheel contact point, in meters.
    - \f$\dot{\theta}\f$ the body angular velocity of the base frame along its
      lateral axis, in radians per seconds.
    - \f$\dot{p}\f$ the velocity of the average wheel contact point, in meters
      per seconds.

    As with all Upkie environments, full observations from the spine (detailed
    in \ref observations) are also available in the `info` dictionary
    returned by the reset and step functions.
    """

    ## \var action_space
    ## Action space.
    action_space: spaces.box.Box

    ## \var leg_return_period
    ## Time constant for the legs (hips and knees) to revert to their neutral
    ## configuration.
    leg_return_period: float

    ## \var left_wheeled
    ## Set to True (default) if the robot is left wheeled, that is, a positive
    ## turn of the left wheel results in forward motion. Set to False for a
    ## right-wheeled variant.
    left_wheeled: bool

    ## \var observation_space
    ## Observation space.
    observation_space: spaces.box.Box

    ## \var reward
    ## Reward function of the environment.
    reward: WheeledInvertedPendulumReward

    ## \var version
    ## Environment version number.
    version = 3

    ## \var wheel_radius
    ## Wheel radius in [m].
    wheel_radius: float

    def __init__(
        self,
        fall_pitch: float = 1.0,
        frequency: float = 200.0,
        frequency_checks: bool = True,
        init_state: Optional[RobotState] = None,
        leg_return_period: float = 1.0,
        left_wheeled: bool = True,
        max_ground_velocity: float = 1.0,
        regulate_frequency: bool = True,
        reward: Optional[WheeledInvertedPendulumReward] = None,
        shm_name: str = "/upkie",
        spine_config: Optional[dict] = None,
        wheel_radius: float = 0.06,
    ):
        r"""!
        Initialize environment.

        \param fall_pitch Fall detection pitch angle, in radians.
        \param frequency Regulated frequency of the control loop, in Hz.
        \param frequency_checks If `regulate_frequency` is set and this
            parameter is true (default), a warning is issued every time the
            control loop runs slower than the desired `frequency`. Set this
            parameter to false to disable these warnings.
        \param init_state Initial state of the robot, only used in simulation.
        \param leg_return_period Time constant for the legs (hips and knees) to
            revert to their neutral configuration.
        \param left_wheeled Set to True (default) if the robot is left wheeled,
            that is, a positive turn of the left wheel results in forward
            motion. Set to False for a right-wheeled variant.
        \param max_ground_velocity Maximum commanded ground velocity in m/s.
        \param regulate_frequency Enables loop frequency regulation.
        \param reward Reward function of the environment.
        \param shm_name Name of shared-memory file.
        \param spine_config Additional spine configuration overriding the
            default `upkie.config.SPINE_CONFIG`. The combined configuration
            dictionary is sent to the spine at every reset.
        \param wheel_radius Wheel radius in [m].
        """
        super().__init__(
            fall_pitch=fall_pitch,
            frequency=frequency,
            frequency_checks=frequency_checks,
            init_state=init_state,
            regulate_frequency=regulate_frequency,
            shm_name=shm_name,
            spine_config=spine_config,
        )

        if self.dt is None:
            raise UpkieException("This environment needs a loop frequency")

        reward: WheeledInvertedPendulumReward = (
            reward if reward is not None else WheeledInvertedPendulumReward()
        )

        # gymnasium.Env: observation_space
        MAX_BASE_PITCH: float = np.pi
        MAX_GROUND_POSITION: float = float("inf")
        MAX_BASE_ANGULAR_VELOCITY: float = 1000.0  # rad/s
        observation_limit = np.array(
            [
                MAX_BASE_PITCH,
                MAX_GROUND_POSITION,
                MAX_BASE_ANGULAR_VELOCITY,
                max_ground_velocity,
            ],
            dtype=float,
        )
        self.observation_space = spaces.Box(
            -observation_limit,
            +observation_limit,
            shape=observation_limit.shape,
            dtype=observation_limit.dtype,
        )

        # gymnasium.Env: action_space
        action_limit = np.array([max_ground_velocity], dtype=float)
        self.action_space = spaces.Box(
            -action_limit,
            +action_limit,
            shape=action_limit.shape,
            dtype=action_limit.dtype,
        )

        self.__leg_servo_action = {
            joint.name: {
                "position": None,
                "velocity": 0.0,
                "maximum_torque": joint.limit.effort,
            }
            for joint in self.model.upper_leg_joints
        }

        self.leg_return_period = leg_return_period
        self.left_wheeled = left_wheeled
        self.reward = reward
        self.wheel_radius = wheel_radius

    def reset(
        self,
        *,
        seed: Optional[int] = None,
        options: Optional[dict] = None,
    ) -> Tuple[np.ndarray, Dict]:
        r"""!
        Resets the environment and get an initial observation.

        \param seed Number used to initialize the environment’s internal random
            number generator.
        \param options Currently unused.
        \return
            - `observation`: Initial vectorized observation, i.e. an element
              of the environment's `observation_space`.
            - `info`: Dictionary with auxiliary diagnostic information. For
              Upkie this is the full observation dictionary sent by the spine.
        """
        return super().reset(seed=seed)

    def parse_first_observation(self, spine_observation: dict) -> None:
        r"""!
        Parse first observation after the spine interface is initialized.

        \param spine_observation First observation.
        """
        for joint in self.model.upper_leg_joints:
            position = spine_observation["servo"][joint.name]["position"]
            self.__leg_servo_action[joint.name]["position"] = position

    def get_env_observation(self, spine_observation: dict) -> np.ndarray:
        r"""!
        Extract environment observation from spine observation dictionary.

        \param spine_observation Spine observation dictionary.
        \return Environment observation vector.
        """
        base_orientation = spine_observation["base_orientation"]
        pitch_base_in_world = base_orientation["pitch"]
        angular_velocity_base_in_base = base_orientation["angular_velocity"]
        ground_position = spine_observation["wheel_odometry"]["position"]
        ground_velocity = spine_observation["wheel_odometry"]["velocity"]

        obs = np.empty(4, dtype=float)
        obs[0] = pitch_base_in_world
        obs[1] = ground_position
        obs[2] = angular_velocity_base_in_base[1]
        obs[3] = ground_velocity
        return obs

    def get_upper_leg_servo_action(self) -> Dict[str, Dict[str, float]]:
        r"""!
        Get servo actions for both hip and knee joints.

        \return Servo action dictionary.
        """
        for joint in self.model.upper_leg_joints:
            prev_position = self.__leg_servo_action[joint.name]["position"]
            new_position = low_pass_filter(
                prev_output=prev_position,
                cutoff_period=self.leg_return_period,
                new_input=0.0,
                dt=self.dt,
            )
            self.__leg_servo_action[joint.name]["position"] = new_position
        return self.__leg_servo_action

    def get_wheel_servo_action(
        self, left_wheel_velocity: float
    ) -> Dict[str, Dict[str, float]]:
        r"""!
        Get servo actions for wheel joints.

        \param[in] left_wheel_velocity Left-wheel velocity, in [rad] / [s].
        \return Servo action dictionary.
        """
        right_wheel_velocity = -left_wheel_velocity
        servo_action = {
            "left_wheel": {
                "position": math.nan,
                "velocity": left_wheel_velocity,
            },
            "right_wheel": {
                "position": math.nan,
                "velocity": right_wheel_velocity,
            },
        }
        for joint in self.model.wheel_joints:
            servo_action[joint.name]["maximum_torque"] = joint.limit.effort
        return servo_action

    def get_spine_action(self, action: np.ndarray) -> dict:
        r"""!
        Convert environment action to a spine action dictionary.

        \param action Environment action.
        \return Spine action dictionary.
        """
        ground_velocity = clamp_and_warn(
            action[0],
            self.action_space.low[0],
            self.action_space.high[0],
            label="ground_velocity",
        )
        wheel_velocity = ground_velocity / self.wheel_radius
        left_wheel_sign = 1.0 if self.left_wheeled else -1.0
        left_wheel_velocity = left_wheel_sign * wheel_velocity
        servo_dict = {}
        servo_dict.update(self.get_upper_leg_servo_action())
        servo_dict.update(self.get_wheel_servo_action(left_wheel_velocity))
        spine_action = {"servo": servo_dict}
        return spine_action

    def get_reward(
        self,
        observation: np.ndarray,
        action: np.ndarray,
    ) -> float:
        r"""!
        Get reward from observation and action.

        \param observation Environment observation vector.
        \param action Environment action vector.
        \return Reward.
        """
        return self.reward(
            pitch=observation[0],
            ground_position=observation[1],
            angular_velocity=observation[2],
            ground_velocity=observation[3],
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2024 Inria

import numpy as np
from gymnasium import Env, Wrapper


class RandomPush(Wrapper):
    """!
    At each step, with a given probability, apply a random push to the robot.
    """

    ## \var env
    ## Wrapped environment.
    env: Env

    # Internal reference to unwrapped get_spine_action function.
    _env_get_spine_action: callable

    ## \var push_generator
    ## Function that generates the push force. It should return a 3D numpy
    ## array.
    push_generator: callable

    ## \var push_prob
    ## Probability of pushing at each step.
    push_prob: float

    def __init__(
        self,
        env,
        push_prob=0.01,
        push_generator=lambda: np.random.normal(0, 400, 3),
    ):
        r"""!
        Initialize wrapper.

        \param env Environment to wrap.
        \param push_prob Probability of pushing at each step.
        \param push_generator Function that generates the push force. It should
            return a 3D numpy array.
        """
        super().__init__(env)
        self.env = env
        self.push_prob = push_prob
        self.push_generator = push_generator
        if not hasattr(self.env, "get_spine_action"):
            raise ValueError(
                "The environment must have a method get_spine_action"
            )
        self._env_get_spine_action = self.env.get_spine_action
        self.env.unwrapped.get_spine_action = self.get_spine_action

    def get_spine_action(self, action):
        """Adds a random push to the action."""
        spine_action = self._env_get_spine_action(action)
        if np.random.binomial(1, self.push_prob):
            force = self.push_generator()
            spine_action["bullet"] = {
                "external_forces": {"torso": {"force": force}}
            }
        else:  # Reset the forces to zero
            spine_action["bullet"] = {
                "external_forces": {"torso": {"force": np.zeros(3)}}
            }
        return spine_action
